/*
	Header file for the hCount algorithm.
	Implementation - hC.c
	
	As described in  - 
	@inproceedings{refHC,
		author = {Cheqing Jin and Weining Qian and Chaofeng Sha and Jeffrey X. Yu and Aoying Zhou},
		title = {Dynamically maintaining frequent items over a data stream},
		booktitle = {CIKM '03: Proceedings of the twelfth international conference on Information and knowledge management},
		year = {2003},
		isbn = {1-58113-723-0},
		pages = {287--294},
		location = {New Orleans, LA, USA},
		doi = {http://doi.acm.org/10.1145/956863.956918},
		publisher = {ACM Press},
		address = {New York, NY, USA},
	}
*/

typedef unsigned int S_TYPE;

typedef struct hC_type {
    long N;			/* global count of total no. of items in stream */
    S_TYPE **S;		/* S[m][h] hash tables */
    int m, h;		/* dimensions of the hash table array S */
    int *a, *b;		/* a_i and b_i -> array of h random numbers */
    int P;			/* the prime number used by the hashing function */
    long M;			/* largest number in data stream */
} hC_type;

extern void hC_Print(hC_type * hc);
extern hC_type *hC_Init(float rho, float epsilon, long M);
extern void hC_Update(hC_type *, int, int);
extern unsigned int *hC_Output(hC_type *, int, int);
extern void hC_Destroy(hC_type *);
extern int hC_Size(hC_type *);
