/********************************************************************
Approximate frequent items in a data stream
G. Cormode 2002, 2003

Last modified: 2003-12

*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>		
#include <string.h>		
#include "prng.h"
#include "massdal.h"

/******************************************************************/

#include "cgt.h"
#include "lossycount.h"
#include "frequent.h"
#include "ccfc.h"
#include "countmin.h"
#include "hc.h"
#include "ss.h"

void mainloop(int);
void print_output();
void GetCmdParams(int, char **);
/******************************************************************/

/* macro that returns a random number in [v1, v2) */  
#define random(v1,v2) (((float)(v1))+(((float)(v2)-((float)(v1)))*rand()/(RAND_MAX+1.0f)))

float zipfpar, phi, delparam;
int range;
long rseed;			/* flag which tells whether to initialize RN Generator with clock() or a const. value */
int ToOrder;		/* flag which tells whether to order stream */
int *exact;
int *stream;
int d;
long n = 0xfffff, lgn = 10;		/* Just for ref. all these will get overwritten */
int width = 10, depth = 2, gran = 1;
int netpos;
int quartiles[4], max;
long long sumsq;
char fdata[20];		/* name of data file */
int REALDATA = 0;	/* flag to tell whether we are using data-file or need to generate data */
int RUNS;
char opstr[350][60];	// provision for 50 runs
float delta, epsilon;
/******************************************************************/

int testout(int item, int thresh)
{
    if (exact[item] >= thresh)
		return 1;
    else
		return 0;
}


/******************************************************************/

/*
NDM - Modified to 
(1) Use clock or user provided seed for Random Number Generator
*/
int *CreateStream(int length)
{
    long a, b;
    float zet;
    int i;
    long value;
    int *stream;
    prng_type *prng;

    int x;

    exact = (int *) calloc(n + 1, sizeof(int));
    stream = (int *) calloc(length + 1, sizeof(int));

	/* Use clock or provided seed. Useful when repeating expt. 
	with different params but same stream is required */
    if (rseed == 0) {
		prng = prng_Init(time(NULL), 2);
		srand((unsigned) time(NULL));
    } else {
		prng = prng_Init(rseed, 2);
		srand(rseed);
	}

    a = (long long) (prng_int(prng) % MOD);
    b = (long long) (prng_int(prng) % MOD);

    netpos = 0;

    zet = zeta(length, zipfpar);

    for (i = 1; i <= length; i++) {
		value = (hash31(a, b, ((int) floor(fastzipf(zipfpar, n, zet, prng)))) & n);
		
		exact[value]++;
		netpos++;
		
		stream[i] = value;
    }

    prng_Destroy(prng);
    lgn = 20;

    return (stream);

}

/*
NDM
Read the stream from file fdata. Reads the first n items.
(For use with real data)
*/
int *ReadStream()
{
	char str[12];
	int i;
	int *stream;
	FILE *fin;

	fin=fopen(fdata, "r");
	if (fin==NULL) {
		printf("\nData file error.");
		exit(20);	
	}

    exact = (int *) calloc(n + 1, sizeof(int));
    stream = (int *) calloc(range + 1, sizeof(int));
	if (exact==NULL || stream==NULL) {
		printf("calloc error");
		exit(123);
	}

	i=1;
	while (i < range) {
		fscanf(fin, "%s", str);
		stream[i] = atoi(str);
		stream[i] &= n;
		if (stream[i] >= 0) {
			exact[stream[i]]++;
			netpos++;
		}
		else
			exact[-stream[i]]--;
		i++;
	}
	fflush(stdout);
	return(stream);
}



/******************************************************************/

void CheckOutput(const char *title, unsigned int *resultlist,
		 int thresh, int hh, int upt, int outt, int space)
{
	static int opcnt=0;

    int i;
    int correct = 0;
    int claimed = 0;
    unsigned int last = 0;
    float recall, precision;

    if (resultlist[0] > 0) {
		//tshell(resultlist[0],resultlist);
		//printf("Looking at a list of %d items\n",resultlist[0]);
		for (i = 1; i <= resultlist[0]; i++) {
			claimed++;
			if (resultlist[i] != last) {
				correct += testout(resultlist[i], thresh);
				last = resultlist[i];
			}
		}
    } else
		claimed = 1;
    
	if (hh == 0) {
		sprintf(opstr[opcnt], "%s\t--\t--\t%d\t%d\n", title, space, upt);
		printf("%s\t--\t--\t%d\t%d\n", title, space, upt);
		opcnt++;
    } else {
		recall = 100.0 * ((float) correct) / ((float) hh);
		precision = 100.0 * ((float) correct) / ((float) claimed);
		sprintf(opstr[opcnt], "%s,%1.2f,%1.2f,%d,%d,%d\n", title, recall, precision,
			   space, upt, outt);
		printf("%s,%1.2f,%1.2f,%d,%d,%d\n", title, recall, precision,
			   space, upt, outt);
		opcnt++;
    }
}

/******************************************************************/

int RunExact(int thresh)
{
    int i, hh, sum, j;
    StartTheClock();
    hh = 0;
    sum = -1, j;
	
	j = 0;
    sumsq = 0;
    for (i = 0; i < n; i++) {
		sum += exact[i];
		sumsq += (long long) exact[i] * (long long) exact[i];
		if (exact[i] > 0)
			max = i;
		while (sum >= 0) {
			quartiles[j] = i;
			j++;
			sum -= netpos / 4;
		}
		if (exact[i] >= thresh) {
			hh++;
		}
    }
    
	printf("\n#%d Frequent Items\n", hh);
	printf("Exact,100.00,100.0,%d,0,%d\n", n*sizeof(int), StopTheClock());

	return (hh);
}



/******************************************************************/

int main(int argc, char **argv)
{
	int i;

	GetCmdParams(argc, argv);

	if (d <= 0)
		exit(1);

	printf("\nMethod,Recall,Precis,Space,Utime,Qtime\n");
    
	for (i=0; i<RUNS; i++) {
		printf("\nRUN: %d", i);
		mainloop(i);
	}
	
	print_output();
}

void print_output()
{
	int i, j;

	for (i=0; i<7; i++) {
		printf("\n\n");
		for (j=0; j<RUNS; j++) {
			printf("%s", opstr[j*7+i]);
		}
	}
}

void mainloop(int runnum)
{
	int i, uptime, outtime;
    int thresh, hh;
    unsigned int *uilist;
    		
    CGT_type *cgt;
    LC_type *lc;
    freq_type *freq;
    CCFC_type *ccfc;
    CMH_type *cmh;
    hC_type *hc;
	SS_type *ss;

	if (REALDATA) {
		if (runnum == 0)
			stream = ReadStream();
	} else {
		stream = CreateStream(range);
	}

    thresh = floor(phi * netpos);
    if (thresh == 0)
		thresh = 1;
    
    hh = RunExact(thresh);
	
	cgt = CGT_Init(phi, delta, n, gran);
    StartTheClock();
    for (i = 1; i <= range; i++) {
		if (stream[i] > 0)
			CGT_Update(cgt, stream[i], 1);
		else
			CGT_Update(cgt, -stream[i], -1);
		}
    uptime = StopTheClock();
    StartTheClock();
    uilist = CGT_Output(cgt, thresh);
    outtime = StopTheClock();
    CheckOutput("CGT", uilist, thresh, hh, uptime, outtime, CGT_Size(cgt));
    free(uilist);
	CGT_Destroy(cgt);

    cmh = CMH_Init(epsilon, delta, n, gran);
    StartTheClock();
    for (i = 1; i <= range; i++)
		if (stream[i] > 0)
			CMH_Update(cmh, stream[i], 1);
		else
			CMH_Update(cmh, -stream[i], -1);
    uptime = StopTheClock();
    StartTheClock();
    uilist = CMH_FindHH(cmh, thresh);
    outtime = StopTheClock();
    CheckOutput("CM", uilist, thresh, hh, uptime, outtime, CMH_Size(cmh));
    free(uilist);
	CMH_Destroy(cmh);	

	ccfc = CCFC_Init(phi, delta, n, gran);
    StartTheClock();
    for (i = 1; i <= range; i++)
		if (stream[i] > 0)
			CCFC_Update(ccfc, stream[i], 1);
		else
			CCFC_Update(ccfc, -stream[i], -1);
    uptime = StopTheClock();
    StartTheClock();
    uilist = CCFC_Output(ccfc, thresh);
    outtime = StopTheClock();
    CheckOutput("CCFC", uilist, thresh, hh, uptime, outtime, CCFC_Size(ccfc));
    free(uilist);
	CCFC_Destroy(ccfc);

    lc = LC_Init(epsilon);
    StartTheClock();
    for (i = 1; i <= range; i++)
		LC_Update(lc, stream[i]);	// checks for negative values (deletions) within the function
    uptime = StopTheClock();
    StartTheClock();
    uilist = LC_Output(lc, thresh);
    outtime = StopTheClock();
    CheckOutput("LC", uilist, thresh, hh, uptime, outtime, LC_Size(lc));
    free(uilist);
	LC_Destroy(lc);
	
    freq = Freq_Init(phi);
    StartTheClock();
    for (i = 1; i <= range; i++)
		Freq_Update(freq, stream[i]);	// checks for negative values (deletions) within the function
    uptime = StopTheClock();
    StartTheClock();
    uilist = Freq_Output(freq, thresh);
    outtime = StopTheClock();
    CheckOutput("Freq", uilist, thresh, hh, uptime, outtime,
		Freq_Size(freq));
    free(uilist);
	Freq_Destroy(freq);

	ss = SS_Init(epsilon);
	StartTheClock();
    for (i = 1; i <= range; i++)
		if (stream[i] > 0)
			SS_Update(ss, stream[i], 1);
		else
			SS_Update(ss, -stream[i], -1);
    uptime = StopTheClock();
	StartTheClock();
    uilist = SS_Output(ss, thresh);
    outtime = StopTheClock();
	CheckOutput("ss", uilist, thresh, hh, uptime, outtime, SS_Size(ss));
	free(uilist);
	SS_Destroy(ss);

    hc = hC_Init(epsilon, delta, n);
    StartTheClock();
    for (i = 1; i <= range; i++)
		if (stream[i] > 0)
			hC_Update(hc, stream[i], 1);
		else
			hC_Update(hc, -stream[i], -1);
    uptime = StopTheClock();
	StartTheClock();
    uilist = hC_Output(hc, thresh, 1);
    outtime = StopTheClock();
    CheckOutput("hC", uilist, thresh, hh, uptime, outtime, hC_Size(hc));
    free(uilist);
	hC_Destroy(hc);
	 	
	if (!REALDATA) {
		free(stream);
		free(exact);
	}
	
}

void GetCmdParams(int argc, char **argv)
{
    range = atoi(argv[1]);
    zipfpar = atof(argv[2]);
    phi = atof(argv[3]);
    delparam = atof(argv[4]);	/* deletion probability - 0 to 1 */
    delta = atof(argv[5]);
    epsilon = atof(argv[6]) * phi;
    gran = atoi(argv[7]);
    rseed = atol(argv[8]);
    n = atol(argv[9]);			/* maximum number in stream */
    ToOrder = atoi(argv[10]);
	RUNS = atoi(argv[11]);
	
	if (argc>12) {
		strcpy(fdata, argv[12]);
		REALDATA = 1;
	}
	

	d = (int) ceil(1.0 / phi);

    printf("\n# %d %f %f %f %f %f %d %ld %ld %d %d\n", range, zipfpar, phi, delparam,
	   delta, epsilon, gran, rseed, n, ToOrder, RUNS);
	printf("\nREALDATA: %d", REALDATA);
	
}

